<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <title>{% block title %}Flask-live-twcss{% endblock %}</title>
  
  {% block head %}{% endblock %}

  {live_reload_template_placeholder}
</head>
<body>
  {% block content %}{% endblock %}
</body>
</html>
