// Docs: https://tailwindcss.com/docs/configuration

/** @type {import('tailwindcss').Config} */
module.exports = {
  content: [
    '{content_glob_placeholder}',
  ],
  theme: {
    extend: {},
  },
  plugins: [],
}
